from datetime import datetime, timezone
from elody.object_configurations.base_object_configuration import (
    BaseObjectConfiguration,
)
from elody.util import flatten_dict
from uuid import uuid4


class ElodyConfiguration(BaseObjectConfiguration):
    SCHEMA_TYPE = "elody"
    SCHEMA_VERSION = 1

    def crud(self):
        crud = {
            "creator": lambda post_body, **kwargs: self._creator(post_body, **kwargs),
            "post_crud_hook": lambda **kwargs: self._post_crud_hook(**kwargs),
            "pre_crud_hook": lambda **kwargs: self._pre_crud_hook(**kwargs),
        }
        return {**super().crud(), **crud}

    def document_info(self):
        return {"object_lists": {"metadata": "key", "relations": "type"}}

    def logging(self, flat_document, **kwargs):
        return super().logging(flat_document, **kwargs)

    def migration(self):
        return super().migration()

    def serialization(self, from_format, to_format):
        return super().serialization(from_format, to_format)

    def validation(self):
        return super().validation()

    def _creator(
        self,
        post_body,
        *,
        get_user_context,
        flat_post_body={},
        document_defaults={},
    ):
        if not flat_post_body:
            flat_post_body = flatten_dict(
                self.document_info()["object_lists"], post_body
            )
        _id = document_defaults.get("_id", str(uuid4()))

        identifiers = []
        for property in self.document_info().get("identifier_properties", []):
            if identifier := flat_post_body.get(f"metadata.{property}.value"):
                identifiers.append(identifier)

        template = {
            "_id": _id,
            "computed_values": {
                "created_at": datetime.now(timezone.utc),
                "event": "create",
            },
            "identifiers": list(
                set([_id, *identifiers, *document_defaults.pop("identifiers", [])])
            ),
            "metadata": [],
            "relations": [],
            "schema": {"type": self.SCHEMA_TYPE, "version": self.SCHEMA_VERSION},
        }
        if email := self.__get_email(get_user_context):
            template["computed_values"]["created_by"] = email

        for key in self.document_info()["object_lists"].keys():
            post_body = self._get_merged_post_body(post_body, document_defaults, key)
        document = {**template, **document_defaults, **post_body}

        self._sanitize_document(document, "metadata", "value")
        self._sort_document_keys(document)
        return document

    def _post_crud_hook(self, **_):
        pass

    def _pre_crud_hook(self, *, crud, document={}, get_user_context=None, **_):
        if document:
            self._sanitize_document(document, "metadata", "value")
            self.__patch_document_computed_values(
                crud, document, get_user_context=get_user_context
            )
            self._sort_document_keys(document)

    def __get_email(self, get_user_context):
        try:
            return get_user_context().email
        except Exception:
            return None

    def __patch_document_computed_values(self, crud, document, **kwargs):
        document["computed_values"].update({"event": crud})
        document["computed_values"].update({"modified_at": datetime.now(timezone.utc)})
        if email := self.__get_email(kwargs.get("get_user_context")):
            document["computed_values"].update({"modified_by": email})



import re as regex

from configuration import get_object_configuration_mapper  # pyright: ignore
from copy import deepcopy
from elody.util import flatten_dict, interpret_flat_key
from inuits_policy_based_auth.contexts.user_context import UserContext
from logging_elody.log import log  # pyright: ignore


_permissions = {}


def set_permissions(permissions: dict):
    global _permissions
    _permissions = permissions


def get_permissions(role: str, user_context: UserContext):
    permissions = deepcopy(_permissions)
    placeholders = ["X_TENANT_ID", "TENANT_DEFINING_ENTITY_ID"]

    for placeholder in placeholders:
        permissions = __replace_permission_placeholders(
            permissions, placeholder, user_context.bag[placeholder.lower()]
        )
    return permissions.get(role, {})  # pyright: ignore


def __replace_permission_placeholders(data, placeholder_key, placeholder_value):
    if isinstance(data, dict):
        for key, value in data.items():
            data[key] = __replace_permission_placeholders(
                value, placeholder_key, placeholder_value
            )
    elif isinstance(data, list):
        data = [
            __replace_permission_placeholders(item, placeholder_key, placeholder_value)
            for item in data
        ]
    elif isinstance(data, str):
        data = data.replace(placeholder_key, placeholder_value)
    return data


def handle_single_item_request(
    user_context: UserContext, item, permissions, crud, request_body: dict = {}
):
    try:
        item_in_storage_format, flat_item, object_lists, restrictions_schema = (
            __prepare_item_for_permission_check(item, permissions, crud)
        )

        is_allowed_to_crud_item = (
            __is_allowed_to_crud_item(flat_item, restrictions_schema)
            if flat_item
            else None
        )
        if not is_allowed_to_crud_item:
            return is_allowed_to_crud_item

        return __is_allowed_to_crud_item_keys(
            user_context,
            item_in_storage_format,
            flat_item,
            restrictions_schema,
            crud,
            object_lists,
            flatten_dict(object_lists, request_body),
        )
    except Exception as exception:
        log.debug(
            f"{exception.__class__.__name__}: {str(exception)}",
            item.get("storage_format", item),
        )
        if crud != "read":
            log.debug(f"Request body: {request_body}", {})
        raise exception


def mask_protected_content_post_request_hook(user_context: UserContext, permissions):
    def __post_request_hook(response):
        items = response["results"]
        for item in items:
            try:
                (
                    item_in_storage_format,
                    flat_item,
                    object_lists,
                    restrictions_schema,
                ) = __prepare_item_for_permission_check(item, permissions, "read")
                if not flat_item:
                    continue

                __is_allowed_to_crud_item_keys(
                    user_context,
                    item_in_storage_format,
                    flat_item,
                    restrictions_schema,
                    "read",
                    object_lists,
                )
            except Exception as exception:
                log.debug(
                    f"{exception.__class__.__name__}: {str(exception)}",
                    item.get("storage_format", item),
                )
                raise exception

        return response

    return __post_request_hook


def __prepare_item_for_permission_check(item, permissions, crud):
    item = deepcopy(item.get("storage_format", item))
    if item.get("type", "") not in permissions[crud].keys():
        return item, None, None, None

    config = get_object_configuration_mapper().get(item["type"])
    object_lists = config.document_info()["object_lists"]
    flat_item = flatten_dict(object_lists, item)

    return (
        item,
        flat_item,
        object_lists,
        __get_restrictions_schema(flat_item, permissions, crud),
    )


def __get_restrictions_schema(flat_item, permissions, crud):
    schema_type = flat_item.get("schema.type", "elody")
    schema_version = flat_item.get("schema.version", "1")
    schema = f"{schema_type}:{schema_version}"

    schemas = permissions[crud][flat_item["type"]]
    if restrictions_schema := schemas.get(schema):
        return restrictions_schema

    for schema in reversed(schemas.keys()):
        if regex.match(f"^{schema_type}:[0-9]{1,3}?$", schema):
            break
        schema = None
    return schemas[schema] if schemas and schema else {}


def __is_allowed_to_crud_item(flat_item, restrictions_schema):
    restrictions = restrictions_schema.get("object_restrictions", {})

    for restricted_key, restricting_values in restrictions.items():
        restricted_key = restricted_key.split(":")[1]
        item_value_in_restricting_values = __item_value_in_values(
            flat_item, restricted_key, restricting_values
        )
        if not item_value_in_restricting_values:
            return None

    return True


def __is_allowed_to_crud_item_keys(
    user_context: UserContext,
    item_in_storage_format,
    flat_item,
    restrictions_schema,
    crud,
    object_lists,
    flat_request_body: dict = {},
):
    user_context.bag["soft_call_response_body"] = []
    restrictions = restrictions_schema.get("key_restrictions", {})

    for restricted_key, restricting_conditions in restrictions.items():
        restricted_key = restricted_key.split(":")[1]
        condition_match = True
        for condition_key, condition_values in restricting_conditions.items():
            condition_match = __item_value_in_values(
                flat_item, condition_key, condition_values, flat_request_body
            )
            if not condition_match:
                break

        if condition_match:
            if crud == "read":
                keys_info = interpret_flat_key(restricted_key, object_lists)
                element = item_in_storage_format
                for info in keys_info:
                    if info["is_object_list"]:
                        element = __get_element_from_object_list_of_item(
                            element,
                            info["key"],
                            info["object_key"],
                            object_lists,
                        )
                element[info["key"]] = "[protected content]"  # pyright: ignore
            else:
                if flat_request_body.get(restricted_key):
                    user_context.bag["soft_call_response_body"].append(restricted_key)

    user_context.bag["requested_item"] = item_in_storage_format
    return len(user_context.bag["soft_call_response_body"]) == 0


def __item_value_in_values(flat_item, key, values: list, flat_request_body: dict = {}):
    negate_condition = False
    if key[0] == "!":
        key = key[1:]
        negate_condition = True

    item_value = flat_request_body.get(key, flat_item[key])
    expected_values = []
    for value in values:
        if flat_item_key_value := flat_item.get(value):
            value = flat_item_key_value
        if isinstance(value, list):
            expected_values.extend(value)
        else:
            expected_values.append(value)

    if isinstance(item_value, (str, int, float, bool)):
        if negate_condition:
            return item_value not in expected_values
        else:
            return item_value in expected_values
    elif isinstance(item_value, list):
        for expected_value in expected_values:
            if expected_value in item_value:
                return True != negate_condition
        return False != negate_condition

    raise Exception(f"Invalid item_value: {item_value}")


def __get_element_from_object_list_of_item(
    item: dict, object_list: str, key: str, object_lists: dict
):
    for element in item[object_list]:
        if element[object_lists[object_list]] == key:
            return element
    return {}



from flask_restful import abort  # pyright: ignore


def get_item(storage_manager, user_context_bag, view_args):
    view_args = view_args or {}
    id = view_args.get("id")
    resolve_collections = user_context_bag["collection_resolver"]
    collections = resolve_collections(collection=view_args.get("collection"), id=id)
    for collection in collections:
        if item := storage_manager.get_db_engine().get_item_from_collection_by_id(
            collection, id
        ):
            return item
    else:
        abort(404, message=f"Item with id {id} does not exist.")



import re as regex

from copy import deepcopy
from elody.policies.permission_handler import (
    get_permissions,
    mask_protected_content_post_request_hook,
)
from flask import Request  # pyright: ignore
from inuits_policy_based_auth import BaseAuthorizationPolicy  # pyright: ignore
from inuits_policy_based_auth.contexts.policy_context import (  # pyright: ignore
    PolicyContext,
)
from inuits_policy_based_auth.contexts.user_context import (  # pyright: ignore
    UserContext,
)


class FilterGenericObjectsPolicy(BaseAuthorizationPolicy):
    def authorize(
        self, policy_context: PolicyContext, user_context: UserContext, request_context
    ):
        request: Request = request_context.http_request
        if not regex.match("^(/[^/]+/v[0-9]+)?/[^/]+/filter$", request.path):
            return policy_context

        if not isinstance(user_context.access_restrictions.filters, list):
            user_context.access_restrictions.filters = []
        type_filter = self.__get_type_filter(user_context, request.json or [])
        if not type_filter:
            policy_context.access_verdict = True
            return policy_context

        policy_context.access_verdict = False
        for role in user_context.x_tenant.roles:
            permissions = get_permissions(role, user_context)
            if not permissions:
                continue

            rules = [PostRequestRules]
            access_verdict = None
            for rule in rules:
                access_verdict = rule().apply(
                    type_filter["value"], user_context, request, permissions
                )
                if access_verdict != None:
                    policy_context.access_verdict = access_verdict
                    if not policy_context.access_verdict:
                        return policy_context

            if policy_context.access_verdict:
                return policy_context

        return policy_context

    def __get_type_filter(self, user_context: UserContext, request_body: list):
        type_filter = None
        for filter in request_body:
            if filter["type"] == "type":
                type_filter = filter
            elif (
                filter["type"] == "selection"
                and filter.get("parent_key", "") == ""
                and filter["key"] == "type"
            ):
                type_filter = filter

        if not type_filter:
            user_context.access_restrictions.filters.append(  # pyright: ignore
                {
                    "type": "selection",
                    "key": "type",
                    # TODO refactor this in a more generic way
                    "value": [
                        "language",
                        "type",
                        "collectionForm",
                        "institution",
                        "tag",
                        "triple",
                        "person",
                        "externalRecord",
                        "verzameling",
                        "arches_record",
                        "photographer",
                        "creator",
                        "assetPart",
                        "set",
                        "license",
                        "mediafile",
                        "share_link",
                    ],
                    "match_exact": True,
                }
            )
            return None
        return type_filter


class PostRequestRules:
    def apply(
        self,
        type_filter_values: str | list[str],
        user_context: UserContext,
        request: Request,
        permissions,
    ) -> bool | None:
        if request.method != "POST":
            return None

        if isinstance(type_filter_values, str):
            type_filter_values = [type_filter_values]

        type_filter_values_copy = deepcopy(type_filter_values)
        for type_filter_value in type_filter_values_copy:
            if type_filter_value not in permissions["read"].keys():
                type_filter_values.remove(type_filter_value)
                continue

            restrictions_grouped_by_index = {}
            schemas = permissions["read"][type_filter_value]
            for schema in schemas.keys():
                restrictions = schemas[schema].get("object_restrictions", {})
                for restricted_key, restricting_value in restrictions.items():
                    index, restricted_key = restricted_key.split(":")
                    key = f"{schema}|{restricted_key}"
                    if group := restrictions_grouped_by_index.get(index):
                        group["key"].append(key)
                    else:
                        restrictions_grouped_by_index.update(
                            {
                                index: {
                                    "key": [key],
                                    "value": restricting_value,
                                }
                            }
                        )

            for restriction in restrictions_grouped_by_index.values():
                user_context.access_restrictions.filters.append(  # pyright: ignore
                    {
                        "type": "selection",
                        "key": restriction["key"],
                        "value": restriction["value"],
                        "match_exact": True,
                    }
                )

        if len(type_filter_values) == 0:
            return False
        user_context.access_restrictions.post_request_hook = (
            mask_protected_content_post_request_hook(user_context, permissions)
        )
        return True
